@@ -5,14 +5,13 @@ import urllib
 import urlparse
 import re
 import textwrap
-from Cookie import SimpleCookie
+from Cookie import BaseCookie
 from rfc822 import parsedate_tz, mktime_tz, formatdate
 from datetime import datetime, date, timedelta, tzinfo
 import time
 import calendar
 import tempfile
 import warnings
-from webob.util.dictmixin import DictMixin
 from webob.datastruct import EnvironHeaders
 from webob.multidict import MultiDict, UnicodeMultiDict, NestedMultiDict, NoVars
 from webob.etag import AnyETag, NoETag, ETagMatcher, IfRange, NoIfRange
@@ -43,12 +42,12 @@ UTC = _UTC()
 
 def html_escape(s):
     """HTML-escape a string or object
-    
+
     This converts any non-string objects passed into it to strings
     (actually, using ``unicode()``).  All values returned are
     non-unicode strings (using ``&#num;`` entities for all non-ASCII
     characters).
-    
+
     None is treated specially, and returns the empty string.
     """
     if s is None:
@@ -168,6 +167,9 @@ class header_getter(object):
             if self.header in obj.headers:
                 del obj.headers[self.header]
         else:
+            if isinstance(value, unicode):
+                # This is the standard encoding for headers:
+                value = value.encode('ISO-8859-1')
             obj.headers[self.header] = value
 
     def __delete__(self, obj):
@@ -275,6 +277,9 @@ def _parse_date(value):
     if t is None:
         # Could not parse
         return None
+    if t[-1] is None:
+        # No timezone given.  None would mean local time, but we'll force UTC
+        t = t[:9] + (0,)
     t = mktime_tz(t)
     return datetime.fromtimestamp(t, UTC)
 
@@ -481,7 +486,7 @@ class Request(object):
             self.__dict__['charset'] = charset
         if unicode_errors is not NoDefault:
             self.__dict__['unicode_errors'] = unicode_errors
-        if decode_param_names is NoDefault:
+        if decode_param_names is not NoDefault:
             self.__dict__['decode_param_names'] = decode_param_names
 
     def __setattr__(self, attr, value, DEFAULT=[]):
@@ -699,8 +704,10 @@ class Request(object):
 
     def _urlvars__get(self):
         """
-        Return any variables matched in the URL (e.g.,
-        ``wsgiorg.routing_args``).
+        Return any *named* variables matched in the URL.
+
+        Takes values from ``environ['wsgiorg.routing_args']``.
+        Systems like ``routes`` set this value.
         """
         if 'paste.urlvars' in self.environ:
             return self.environ['paste.urlvars']
@@ -712,19 +719,69 @@ class Request(object):
             return result
 
     def _urlvars__set(self, value):
-        del self.urlvars
-        self.environ['wsgiorg.routing_args'] = ((), value)
+        environ = self.environ
+        if 'wsgiorg.routing_args' in environ:
+            environ['wsgiorg.routing_args'] = (environ['wsgiorg.routing_args'][0], value)
+            if 'paste.urlvars' in environ:
+                del environ['paste.urlvars']
+        elif 'paste.urlvars' in environ:
+            environ['paste.urlvars'] = value
+        else:
+            environ['wsgiorg.routing_args'] = ((), value)
 
     def _urlvars__del(self):
         if 'paste.urlvars' in self.environ:
             del self.environ['paste.urlvars']
         if 'wsgiorg.routing_args' in self.environ:
-            del self.environ['wsgiorg.routing_args']
-            
+            if not self.environ['wsgiorg.routing_args'][0]:
+                del self.environ['wsgiorg.routing_args']
+            else:
+                self.environ['wsgiorg.routing_args'] = (self.environ['wsgiorg.routing_args'][0], {})
+
     urlvars = property(_urlvars__get, _urlvars__set, _urlvars__del, doc=_urlvars__get.__doc__)
 
+    def _urlargs__get(self):
+        """
+        Return any *positional* variables matched in the URL.
+
+        Takes values from ``environ['wsgiorg.routing_args']``.
+        Systems like ``routes`` set this value.
+        """
+        if 'wsgiorg.routing_args' in self.environ:
+            return self.environ['wsgiorg.routing_args'][0]
+        else:
+            # Since you can't update this value in-place, we don't need
+            # to set the key in the environment
+            return ()
+
+    def _urlargs__set(self, value):
+        environ = self.environ
+        if 'paste.urlvars' in environ:
+            # Some overlap between this and wsgiorg.routing_args; we need
+            # wsgiorg.routing_args to make this work
+            routing_args = (value, environ.pop('paste.urlvars'))
+        elif 'wsgiorg.routing_args' in environ:
+            routing_args = (value, environ['wsgiorg.routing_args'][1])
+        else:
+            routing_args = (value, {})
+        environ['wsgiorg.routing_args'] = routing_args
+
+    def _urlargs__del(self):
+        if 'wsgiorg.routing_args' in self.environ:
+            if not self.environ['wsgiorg.routing_args'][1]:
+                del self.environ['wsgiorg.routing_args']
+            else:
+                self.environ['wsgiorg.routing_args'] = ((), self.environ['wsgiorg.routing_args'][1])
+
+    urlargs = property(_urlargs__get, _urlargs__set, _urlargs__del, _urlargs__get.__doc__)
+
     def is_xhr(self):
-        """Returns a boolean if X-Requested-With is present and ``XMLHttpRequest``"""
+        """Returns a boolean if X-Requested-With is present and ``XMLHttpRequest``
+
+        Note: this isn't set by every XMLHttpRequest request, it is
+        only set if you are using a Javascript library that sets it
+        (or you set the header yourself manually).  Currently
+        Prototype and jQuery are known to set this header."""
         return self.environ.get('HTTP_X_REQUESTED_WITH', '') == 'XMLHttpRequest'
     is_xhr = property(is_xhr, doc=is_xhr.__doc__)
 
@@ -866,7 +923,7 @@ class Request(object):
 
     str_queryvars = deprecated_property(str_GET, 'str_queryvars',
                                         'use str_GET instead')
-                                        
+
 
     def GET(self):
         """
@@ -906,6 +963,8 @@ class Request(object):
 
     params = property(params, doc=params.__doc__)
 
+    _rx_quotes = re.compile('"(.*)"')
+
     def str_cookies(self):
         """
         Return a *plain* dictionary of cookies as found in the request.
@@ -918,10 +977,14 @@ class Request(object):
                 return vars
         vars = {}
         if source:
-            cookies = SimpleCookie()
+            cookies = BaseCookie()
             cookies.load(source)
             for name in cookies:
-                vars[name] = cookies[name].value
+                value = cookies[name].value
+                unquote_match = self._rx_quotes.match(value)
+                if unquote_match is not None:
+                    value = unquote_match.group(1)
+                vars[name] = value
         env['webob._parsed_cookies'] = (vars, source)
         return vars
 
@@ -1163,10 +1226,12 @@ class Request(object):
         else:
             status, headers, app_iter = self.call_application(
                 application, catch_exc_info=False)
-        return self.ResponseClass(status, headers, app_iter=app_iter, request=self)
+        return self.ResponseClass(
+            status=status, headerlist=headers, app_iter=app_iter,
+            request=self)
 
     #@classmethod
-    def blank(cls, path, environ=None, base_url=None, headers=None):
+    def blank(cls, path, environ=None, base_url=None, headers=None, **kw):
         """
         Create a blank request environ (and Request wrapper) with the
         given path (path should be urlencoded), and any keys from
@@ -1179,6 +1244,9 @@ class Request(object):
         values you pass in will take precedence.  If you pass in
         base_url then wsgi.url_scheme, HTTP_HOST, and SCRIPT_NAME will
         be filled in from that value.
+
+        Any extra keyword will be passed to ``__init__`` (e.g.,
+        ``decode_param_names``).
         """
         if _SCHEME_RE.search(path):
             scheme, netloc, path, qs, fragment = urlparse.urlsplit(path)
@@ -1245,7 +1313,7 @@ class Request(object):
                 env['SCRIPT_NAME'] = urllib.unquote(path)
         if environ:
             env.update(environ)
-        obj = cls(env)
+        obj = cls(env, **kw)
         if headers is not None:
             obj.headers.update(headers)
         return obj
@@ -1258,10 +1326,11 @@ class Response(object):
     Represents a WSGI response
     """
 
-    default_content_type = None
+    default_content_type = 'text/html'
+    default_charset = 'UTF-8'
     default_conditional_response = False
 
-    def __init__(self, status='200 OK', headerlist=None, body=None, app_iter=None,
+    def __init__(self, body=None, status='200 OK', headerlist=None, app_iter=None,
                  request=None, content_type=None, conditional_response=NoDefault,
                  **kw):
         if app_iter is None:
@@ -1270,12 +1339,11 @@ class Response(object):
         elif body is not None:
             raise TypeError(
                 "You may only give one of the body and app_iter arguments")
-        self._app_iter = app_iter
-        self._body = body
         self.status = status
         if headerlist is None:
-            headerlist = []
-        self._headerlist = headerlist
+            self._headerlist = []
+        else:
+            self._headerlist = headerlist
         self._headers = None
         if request is not None:
             if hasattr(request, 'environ'):
@@ -1286,11 +1354,9 @@ class Response(object):
                 self._request = None
         else:
             self._environ = self._request = None
-        if self._body is not None:
-            self.content_length = len(self._body)
         if content_type is not None:
             self.content_type = content_type
-        elif self.default_content_type is not None:
+        elif self.default_content_type is not None and headerlist is None:
             self.content_type = self.default_content_type
         if conditional_response is NoDefault:
             self.conditional_response = self.default_conditional_response
@@ -1298,7 +1364,23 @@ class Response(object):
             self.conditional_response = conditional_response
         if 'charset' in kw:
             # We set this early, so something like unicode_body works later
-            setattr(self, 'charset', kw.pop('charset'))
+            value = kw.pop('charset')
+            if value:
+                self.charset = value
+        elif self.default_charset and not self.charset and headerlist is None:
+            ct = self.content_type
+            if ct and (ct.startswith('text/') or ct.startswith('application/xml')
+                       or (ct.startswith('application/') and ct.endswith('+xml'))):
+                self.charset = self.default_charset
+        if app_iter is not None:
+            self._app_iter = app_iter
+            self._body = None
+        else:
+            if isinstance(body, unicode):
+                self.unicode_body = body
+            else:
+                self.body = body
+            self._app_iter = None
         for name, value in kw.items():
             if not hasattr(self.__class__, name):
                 # Not a basic attribute
@@ -1452,7 +1534,7 @@ class Response(object):
         for match in _PARAM_RE.finditer(params):
             result[match.group(1)] = match.group(2) or match.group(3) or ''
         return result
-        
+
     def _content_type_params__set(self, value_dict):
         if not value_dict:
             del self.content_type_params
@@ -1539,6 +1621,12 @@ class Response(object):
 
     body_file = property(_body_file__get, fdel=_body_file__del, doc=_body_file__get.__doc__)
 
+    def write(self, text):
+        if isinstance(text, unicode):
+            self.unicode_body += text
+        else:
+            self.body += text
+
     def _unicode_body__get(self):
         """
         Get/set the unicode value of the body (using the charset of the Content-Type)
@@ -1593,12 +1681,22 @@ class Response(object):
 
     def set_cookie(self, key, value='', max_age=None,
                    path='/', domain=None, secure=None, httponly=False,
-                   version=None, comment=None):
+                   version=None, comment=None, expires=None):
         """
         Set (add) a cookie for the response
         """
-        cookies = SimpleCookie()
+        if isinstance(value, unicode) and self.charset is not None:
+            value = '"%s"' % value.encode(self.charset)
+        cookies = BaseCookie()
         cookies[key] = value
+        if isinstance(max_age, timedelta):
+            max_age = timedelta.seconds + timedelta.days*24*60*60
+        if max_age is not None and expires is None:
+            expires = datetime.utcnow() + timedelta(seconds=max_age)
+        if isinstance(expires, timedelta):
+            expires = datetime.utcnow() + expires
+        if isinstance(expires, datetime):
+            expires = '"'+_serialize_date(expires)+'"'
         for var_name, var_value in [
             ('max_age', max_age),
             ('path', path),
@@ -1607,6 +1705,7 @@ class Response(object):
             ('HttpOnly', httponly),
             ('version', version),
             ('comment', comment),
+            ('expires', expires),
             ]:
             if var_value is not None and var_value is not False:
                 cookies[key][var_name.replace('_', '-')] = str(var_value)
@@ -1622,7 +1721,7 @@ class Response(object):
         that it should expire immediately.
         """
         self.set_cookie(key, '', path=path, domain=domain,
-                        max_age=0)
+                        max_age=0, expires=timedelta(days=-5))
 
     def unset_cookie(self, key):
         """
@@ -1638,12 +1737,12 @@ class Response(object):
         del self.headers['Set-Cookie']
         found = False
         for header in existing:
-            cookies = SimpleCookie()
+            cookies = BaseCookie()
             cookies.load(header)
             if key in cookies:
                 found = True
                 del cookies[key]
-            header = cookies.output(header='').lstrip()
+                header = cookies.output(header='').lstrip()
             if header:
                 self.headers.add('Set-Cookie', header)
         if not found:
@@ -1984,8 +2083,8 @@ class Response(object):
         if hasattr(app_iter, 'app_iter_range'):
             return app_iter.app_iter_range(start, stop)
         return AppIterRange(app_iter, start, stop)
-        
-        
+
+
 Request.ResponseClass = Response
 Response.RequestClass = Request
 
@@ -2109,7 +2208,7 @@ class ResponseBodyFile(object):
     def writelines(self, seq):
         for item in seq:
             self.write(item)
-        
+
     closed = False
 
     def encoding(self):
@@ -2176,4 +2275,4 @@ class AppIterRange(object):
             return chunk[:-extra]
         self._served += len(chunk)
         return chunk
-            
+
