@@ -32,7 +32,7 @@ class exists_property(object):
         else:
             if self.prop in obj.properties:
                 del obj.properties[self.prop]
-    def __del__(self, obj):
+    def __delete__(self, obj):
         self.__set__(obj, False)
 
 class value_property(object):
@@ -69,7 +69,7 @@ class value_property(object):
             obj.properties[self.prop] = None # Empty value, but present
         else:
             obj.properties[self.prop] = value
-    def __del__(self, obj):
+    def __delete__(self, obj):
         if self.prop in obj.properties:
             del obj.properties[self.prop]
 
@@ -150,6 +150,8 @@ class CacheControl(object):
         return self.__class__(self.properties.copy(), type=self.type)
 
 def serialize_cache_control(properties):
+    if isinstance(properties, CacheControl):
+        properties = properties.properties
     parts = []
     for name, value in sorted(properties.items()):
         if value is None:
